window.Gleam.initSearch([{"doc":"bright","title":"bright","content":"# Bright\n\nBright is a library to help you manage your `model` and your `update` function in\na [Lustre](https://lustre.build) application. As you probably know, in Lustre,\nyour model is the only mutable place of the application, and centralize every\ndata your application uses. If you&#39;re coming from the JS world, it can seems\nweird at first, because it&#39;s usual to have multiple storage places in an\napplication, whether they&#39;re contexts, stores, observables, or anything else.\n\nIn such a centralized model, everything is simpler, and just works. However,\nmanaging your model and your updates can quickly become a mess, with dependent\ndata, data splitting, normalization, etc. Bright comes in to help you avoid such\nstates, by both helping your to maintain a properly defined model, but also\nto define dependant data in an easy way. Bright also provides a powerful caching\nsystem to guarantee to not compute the same information twice!\n\n&gt; As usual, a demo is worth a thousand words, so take a look at\n&gt; [https://bright.chouquette.dev](https://bright.chouquette.dev)!\n\nIf you&#39;re used to stores &amp; data management, you can skip the next section, see\nyou in the [getting started](#getting-started)!\n\n## Installation\n\n```sh\ngleam add bright@1\n```\n\n## Dependent data? Caching system?\n\nTo sum up simply, a data is dependent on another data when the latter data is\nrequired to compute the former. If you have worked with relational data, you\nalready encountered it. Let&#39;s take an example.\n\nImagine you have a list of users on one side, with their info (name, age for example)\nand ID, and you have a list of addresses referring to a user by their ID. You want to\ncreate a page displaying the user info and all of their addresses.\nIntuitively, you would display the user info, and then you would find all the user\naddresses in the address list to display them.\n\n```gleam\npub type User {\n  User(\n    name: String,\n    age: Int,\n    id: String\n  )\n}\n\npub type Address {\n  Address(\n    street: String,\n    city: String,\n    country: String,\n    user_id: String\n  )\n}\n\npub fn display_user_page(model, user_id) {\n  // First find the user.\n  use user &lt;- result.try(list.find(model.users, fn (user) { user.id == user_id }))\n  // After the user has been found, filter the addresses to find them.\n  use addresses &lt;- list.filter(model.addresses, fn (address) { address.user_id == user.id })\n  // Do the display here.\n}\n```\n\nWhile it works perfectly in this example, what if every time you need to access\nthose data? Recompute the same data over and over again, in every part of the\napplication? We can do better: define the user and its address in one record,\nand use it everywhere!\n\n```gleam\npub type UserAndAddress {\n  UserAndAddress(\n    user: User,\n    addresses: List(Address)\n  )\n}\n\n// Compute the data, and store it in your model.\npub fn compute_user_address(model) {\n  list.map(model.users, fn (user) {\n    let addresses = list.filter(model.addresses, fn (address) { address.user_id == user.id })\n    UserAndAddresses(user:, addresses:)\n  })\n}\n\n// The data is in your model.\npub fn display_user_page(model, user_id) {\n  // Find directly everything!\n  use UserAndAddress(user:, addresses:) &lt;- result.try({\n    list.find(model.users_and_addresses, fn (user) { user.id == user_id })\n  })\n  // Do the display here.\n}\n```\n\nYou have defined here a derived, computed data that depends on two previous data\nyou own. Now, every time you need to access the user, you have the address bundled!\nBut in a classic application, you would have to define that computation by yourself,\nand make sure to keep it in sync after each update. That&#39;s where Bright comes in,\nand do the hard work for you! Instead of having to think to synchronize the data\nwhen a new data comes in, let Bright does it for you. And with built-in caching\non-demand, Bright will never recompute the same data twice for intense computations.\n\n## Getting Started\n\nBright handles the hard task of computing the derived data when needed, and as\nsuch, you have to initialize it at first. Bright accepts two types of data:\nyour main model, holding the raw data, and your derived, computed data. Because\nGleam is strongly-typed language, you&#39;ll have to define those data by hand.\nA counter will be used to illustrate how to use it.\n\n```gleam\n/// Define your raw data. No derived data will reside here.\npub type Data {\n  Data(counter: Int)\n}\n\n/// Define your derived data. No raw data will reside here.\npub type Computed {\n  Computed(double: Int)\n}\n\n/// Define an alias, to simplify reference to the model.\npub type Model =\n  Bright(Data, Computed)\n\n// In your init function, you&#39;ll initialize Bright. You can use it as-is as a\n// replacement for your model.\npub fn init() {\n  let data = Data(counter: 0)\n  let computed = Computed(double: 0)\n  let model = bright.init(data, computed)\n  #(model, effect.none())\n}\n```\n\nOnce Bright is initialized, you have to modify a bit your update function. Now,\ninstead of simply receiving the message and modifying your model, you&#39;ll have to\nrun that modification through Bright. You can then chain your computation calls\nto the Bright object. And of course, derived data can be computed from pre-computed\nderived data!\n\n```gleam\npub type Msg {\n  Increment\n  Decrement\n}\n\npub fn update(model: Model, msg: Msg) {\n  // bright.start begins the bright update cycle. It&#39;s required to use bright\n  // in efficiently.\n  // By using function capture, we can easily use our update function here.\n  // bright.update will automatically run your update against data, here our\n  // Data record. Like every update function, that function have to return\n  // a #(Data, Effect(Msg)). The message will automatically be batched with\n  // next messages.\n  // Finally, Bright(Data, Computed) is returned, with Data updated. To let you\n  // continue the chain.\n  use model &lt;- bright.start(model)\n  use model &lt;- bright.update(model, update_data(_, msg))\n  model\n  // bright.compute will compute the new derived data, and let you set it in\n  // the computed. You can also simply return the original computed, in which\n  // case the data is not updated.\n  |&gt; bright.compute(fn(data, computed) { Computed(..computed, double: d.counter * 2) })\n  // bright.lazy_compute will compute the new derived data, if and only if the\n  // selector you pass as the first argument changed between two renders.\n  // In case the selector did not change, the old data is kept in memory for the\n  // next render. When the compute function runs, data, computed and the result\n  // of the selector are provided to the function.\n  |&gt; bright.lazy_compute(\n    // That selector value is compared at every render.\n    fn (data) { data.counter / 10 },\n    // _counter is equal to data.counter / 10.\n    fn(data, computed, _counter) { Computed(..computed, double: d.counter * 2) }\n  )\n}\n\npub fn update_data(data: Data, msg: Msg) {\n  case msg {\n    Increment -&gt; #(Data(..data, counter: data.counter + 1), effect.none())\n    Decrement -&gt; #(Data(..data, counter: data.counter - 1), effect.none())\n  }\n}\n```\n\nAnd once your data is computed, all you have to do is to run through your view\nfunction!\n\n```gleam\npub fn view(model: Model) {\n  let #(data, computed) = bright.unwrap(model)\n  // You can use data &amp; computed with correct, up to date data.\n  html.div([], [])\n}\n```\n\nAnd you&#39;re good to go! Now, you don&#39;t have to think anymore to update your\nderived data, everything is kept in-sync directly for you!\n\n## Scheduling side-effects\n\nSometimes, you also have to define side-effects that run after your computations\nhave run. Because you figure out the data is finally incorrect. Or because your\nuser have written a false URL in the address bar. Bright got you covered too!\nJust use `bright.schedule`, and let the side-effects flow automatically in your\napp, only when you need it!\n\n```gleam\npub fn update(model: Model, msg: Msg) {\n  use model &lt;- bright.update(model, update_data(_, msg))\n  model\n  |&gt; bright.compute(fn(data, computed) { Computed(..computed, double: d.counter * 2) })\n  |&gt; bright.lazy_compute(\n    fn (data) { data.counter / 10 },\n    fn(data, computed) { Computed(..computed, double: d.counter * 2) }\n  )\n  // bright.schedule will run at every render, and let you the possibility to issue\n  // a side-effect. Bright will take care to gather them, and provide them to the\n  // runtime!\n  |&gt; bright.schedule(fn (data, computed) {\n    effect.from(fn (dispatch) {\n      io.println(&quot;That side-effect will run at every render!&quot;)\n    })\n  })\n  // bright.lazy_schedule will issue the side-effect, if and only if the\n  // selector you pass as the first argument changed between two renders.\n  // In case the selector did not change, the old data is kept in memory for the\n  // next render. When the compute function runs, data, computed and the result\n  // of the selector are provided to the function.\n  |&gt; bright.lazy_schedule(\n    fn (data) { data.counter / 10 },\n    // _counter is equal to data.counter / 10.\n    fn (data, computed, _counter) {\n      effect.from(fn (dispatch) {\n        io.println(&quot;That side-effect will only run when the selector changes!&quot;)\n      })\n    }\n  )\n}\n```\n\n## Combining multiple Bright\n\nSometimes, you also need to combine multiple Bright in the same model. While you\ncan keep a `Bright` as model, you could want to combine them, to handle one `Bright`\nby page for example. `bright.step` helps you to do this.\n\n```gleam\npub type Model {\n  Model(\n    counter_1: Bright(Data, Computed),\n    counter_2: Bright(Data, Computed),\n  )\n}\n\npub type Msg {\n  First(counter: Counter)\n  Second(counter: Counter)\n}\n\npub type Counter {\n  Decrement\n  Increment\n}\n\n/// Here, we define a new update function, that calls our previously defined\n/// update function. It keeps the two Bright synchronized by running the full\n/// updated cycle on each of them.\nfn update_both_counters(model: Model, msg: Msg) {\n  use counter_1 &lt;- bright.step(update(model.counter_1, msg.counter))\n  use counter_2 &lt;- bright.step(update(model.counter_2, msg.counter))\n  #(Model(..model, counter_1:, counter_2:), effect.none())\n}\n```\n","url":"index.html"},{"doc":"bright","title":"Bright","content":"pub opaque type Bright(data, computed)\n `Bright` holds raw data and computed data, and is used to compute caching.\n `Bright` is instanciated using `init`, with initial data and computed data.\n\nSynonyms:\nbright.Bright\nbright Bright","url":"bright.html#Bright"},{"doc":"bright","title":"compute","content":"pub fn compute(\n  bright: Bright(a, b),\n  compute_: fn(a, b) -&gt; b,\n) -&gt; Bright(a, b)\n Derives data from the `data` state, and potentially the current `computed`\n state. `compute` will run **at every render**, so be careful with computations\n as they can block paint or actors.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   use model &lt;- bright.start(model)\n   use model &lt;- bright.update(model, update_data(_, msg))\n   model\n   |&gt; bright.compute(fn (d, c) { Computed(..c, field1: computation1(d)) })\n   |&gt; bright.compute(fn (d, c) { Computed(..c, field2: computation2(d)) })\n   |&gt; bright.compute(fn (d, c) { Computed(..c, field3: computation3(d)) })\n }\n \nSynonyms:\nbright.compute\nbright compute","url":"bright.html#compute"},{"doc":"bright","title":"computed","content":"pub fn computed(bright: Bright(a, b)) -&gt; b\n Extracts `computed` state from `Bright`.\n\n \n pub fn view(model: Bright(data, computed)) {\n   let computed = bright.computed(model)\n   html.div([], [\n     // Use computed here.\n   ])\n }\n \nSynonyms:\nbright.computed\nbright computed","url":"bright.html#computed"},{"doc":"bright","title":"data","content":"pub fn data(bright: Bright(a, b)) -&gt; a\n Extracts `data` state from `Bright`.\n\n \n pub fn view(model: Bright(data, computed)) {\n   let data = bright.data(model)\n   html.div([], [\n     // Use data here.\n   ])\n }\n \nSynonyms:\nbright.data\nbright data","url":"bright.html#data"},{"doc":"bright","title":"init","content":"pub fn init(data data: a, computed computed: b) -&gt; Bright(a, b)\n Creates the initial `Bright`. `data` &amp; `computed` should be initialised with\n their correct empty initial state.\nSynonyms:\nbright.init\nbright init","url":"bright.html#init"},{"doc":"bright","title":"lazy_compute","content":"pub fn lazy_compute(\n  bright: Bright(a, b),\n  selector: fn(a) -&gt; c,\n  compute_: fn(a, b, c) -&gt; b,\n) -&gt; Bright(a, b)\n Derives data like [`compute`](#compute) lazily. `lazy_compute` accepts a\n selector as second argument. Each time the selector returns a different data\n than previous run, the computation will run. Otherwise, nothing happens.\n The computation function will receive `data`, `computed` and the selected\n data (i.e. the result from your selector function), in case accessing the\n selected data is needed.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   use model &lt;- bright.start(model)\n   use model &lt;- bright.update(model, update_data(_, msg))\n   model\n   // Here, e is always the result data.field / 10 (the result from selector).\n   |&gt; bright.lazy_compute(selector, fn (d, c, e) { Computed(..c, field1: computation1(d)) })\n   |&gt; bright.lazy_compute(selector, fn (d, c, e) { Computed(..c, field2: computation2(d)) })\n   |&gt; bright.lazy_compute(selector, fn (d, c, e) { Computed(..c, field3: computation3(d)) })\n }\n\n /// Use it with lazy_compute to recompute only when the field when\n /// { old_data.field / 10 } != { data.field / 10 }\n fn selector(d, _) {\n   d.field / 10\n }\n \nSynonyms:\nbright.lazy_compute\nbright lazy_compute","url":"bright.html#lazy_compute"},{"doc":"bright","title":"lazy_schedule","content":"pub fn lazy_schedule(\n  bright: Bright(a, b),\n  selector: fn(a) -&gt; c,\n  schedule_: fn(a, b, c) -&gt; Effect(d),\n) -&gt; Bright(a, b)\n Plugs in existing `data` like [`schedule`](#schedule) lazily. `lazy_schedule` accepts\n a selector as second argument. Each time the selector returns a different data\n than previous run, the computation will run. Otherwise, nothing happens.\n The scheduling function will receive `data`, `computed` and the selected\n data (i.e. the result from your selector function), in case accessing the\n selected data is needed.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   use model &lt;- bright.start(model)\n   use model &lt;- bright.update(model, update_data(_, msg))\n   // e is equal to d.field / 10 (the result from selector).\n   use d, c, e &lt;- bright.lazy_schedule(model, selector)\n   use dispatch &lt;- effect.from\n   case d.field == 10 {\n     True -&gt; dispatch(my_msg)\n     False -&gt; Nil\n   }\n }\n\n /// Use it with lazy_schedule to recompute only when the field when\n /// { old_data.field / 10 } != { data.field / 10 }\n fn selector(d, _) {\n   d.field / 10\n }\n \nSynonyms:\nbright.lazy_schedule\nbright lazy_schedule","url":"bright.html#lazy_schedule"},{"doc":"bright","title":"schedule","content":"pub fn schedule(\n  bright: Bright(a, b),\n  schedule_: fn(a, b) -&gt; Effect(c),\n) -&gt; Bright(a, b)\n Plugs in existing `data` and `computed` state, to issue some side-effects,\n when your application needs to run side-effects depending on the current state.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   use model &lt;- bright.start(model)\n   use model &lt;- bright.update(model, update_data(_, msg))\n   use d, c &lt;- bright.schedule(model)\n   use dispatch &lt;- effect.from\n   case d.field == 10 {\n     True -&gt; dispatch(my_msg)\n     False -&gt; Nil\n   }\n }\n \nSynonyms:\nbright.schedule\nbright schedule","url":"bright.html#schedule"},{"doc":"bright","title":"start","content":"pub fn start(\n  bright: Bright(a, b),\n  next: fn(Bright(a, b)) -&gt; Bright(a, b),\n) -&gt; #(Bright(a, b), Effect(c))\n Start the Bright update cycle. Use it as a way to trigger the start of `Bright`\n computations, and chain them with other `bright` calls.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   // Starts the update cycle, and returns #(Bright(data, computed), Effect(msg)).\n   use model &lt;- bright.start(model)\n   use model &lt;- bright.update(update_data(_, msg))\n   model\n }\n \nSynonyms:\nbright.start\nbright start","url":"bright.html#start"},{"doc":"bright","title":"step","content":"pub fn step(\n  bright: #(Bright(a, b), Effect(c)),\n  next: fn(Bright(a, b)) -&gt; #(d, Effect(c)),\n) -&gt; #(d, Effect(c))\n Allows to run multiple `update` on multiple `Bright` in the same update cycle.\n Every call to step with compute a new `Bright`, and will let you chain the\n steps.\n\n \n pub type Model {\n   Model(\n     fst_bright: Bright(data, computed),\n     snd_bright: Bright(data, computed),\n   )\n }\n\n fn update(model: Model, msg: Msg) {\n   use fst_bright &lt;- bright.step(update_fst(model.fst_bright, msg))\n   use snd_bright &lt;- bright.step(update_snd(model.snd_bright, msg))\n   #(Model(fst_bright:, snd_bright:), effect.none())\n }\n \nSynonyms:\nbright.step\nbright step","url":"bright.html#step"},{"doc":"bright","title":"unwrap","content":"pub fn unwrap(bright: Bright(a, b)) -&gt; #(a, b)\n Extracts `data` &amp; `computed` states from `Bright`.\n\n \n pub fn view(model: Bright(data, computed)) {\n   let #(data, computed) = bright.unwrap(model)\n   html.div([], [\n     // Use data or computed here.\n   ])\n }\n \nSynonyms:\nbright.unwrap\nbright unwrap","url":"bright.html#unwrap"},{"doc":"bright","title":"update","content":"pub fn update(\n  bright: Bright(a, b),\n  update_: fn(a) -&gt; #(a, Effect(c)),\n  next: fn(Bright(a, b)) -&gt; Bright(a, b),\n) -&gt; Bright(a, b)\n Update data &amp; effects during update cycle. Use it a way to trigger the start\n of `Bright` computations, and chain them with other `bright` calls.\n\n \n pub fn update(model: Bright(data, computed), msg: Msg) {\n   use model &lt;- bright.start(model)\n   // Run an update, and returns #(data, Effect(msg)).\n   use model &lt;- bright.update(model, update_data(_, msg))\n   model\n }\n \nSynonyms:\nbright.update\nbright update","url":"bright.html#update"},{"doc":"bright","title":"bright","content":"","url":"bright.html"}]);